// ALISP engine v0.4.3b
// ALISP is dedicated to Enota, the girl i'll never forget
// coded by Ketmar // Vampire Avalon (psyc://ketmar.no-ip.org/~Ketmar)
// Understanding is not required. Only obedience.
//
// This program is free software. It comes without any warranty, to
// the extent permitted by applicable law. You can redistribute it
// and/or modify it under the terms of the Do What The Fuck You Want
// To Public License, Version 2, as published by Sam Hocevar. See
// http://sam.zoy.org/wtfpl/COPYING for more details.
//
{$INCLUDE alisp3def.inc}
unit alisp3DosU;

interface

uses
  Windows, SysUtils, DateUtils, Classes,
  wildMatchU, alisp3U;


procedure ALDefineDosPrimitives ();


var
  al3dosArgV: TStringList;


implementation


function ALPFileList (n: Integer): Integer;
// (file-list pathstr maskstr flags [sort])
//   builds list of files in the specified path with the specified mask.
//   "flags" bits meaning:
//     0: include directories
//     1: include read/only files
//     2: include hidden files
//     3: include system files
//     4: exclude files (only dirs)
var
  ps, ms: string;
  flg: Integer;
  atr: Word;
  doSort: Boolean;
  excFl: Boolean;
  sr: TSearchRec;
  fl: TStringList;
begin
  if not ALNextStrArg('DOS::FILE-LIST', n, ps, result) then exit;
  if not ALNextStrArg('DOS::FILE-LIST', n, ms, result) then exit;
  if not ALNextIntArg('DOS::FILE-LIST', n, flg, result) then exit;
  if not ALIsNil(n) then
  begin
    n := ALCar(n);
    doSort := not ALIsNil(n) and not ALIsFalse(n);
  end
  else doSort := false;

  result := ALCons(ALNil(), ALNil()); ALPushCell(result);

  fl := TStringList.Create();
  try
    fl.CaseSensitive := false; fl.Sorted := doSort;

    atr := faArchive;
    if (flg and $01) <> 0 then atr := atr or faDirectory;
    if (flg and $02) <> 0 then atr := atr or faReadOnly;
    if (flg and $04) <> 0 then atr := atr or faHidden;
    if (flg and $08) <> 0 then atr := atr or faSysFile;
    excFl := ((flg and $10) <> 0);

    ps := StringReplace(ps, '/', '\', [rfReplaceAll]);
    ms := StringReplace(ms, '/', '\', [rfReplaceAll]);
    if ps <> '' then ps := IncludeTrailingPathDelimiter(ps);
    if (ms <> '') and (ms[1] = '\') then Delete(ms, 1, 1);
    ps := ps+ExtractFilePath(ms); ms := ExtractFileName(ms);
    if ms = '' then ms := '*';

    if (atr and faDirectory) <> 0 then
    begin
      // directories
      flg := FindFirst(ps+'*', atr, sr);
      try
        while flg = 0 do
        begin
          if (sr.attr and faDirectory) <> 0 then
          begin
            if (sr.Name <> '.') and (sr.Name <> '..') then fl.Append(sr.name+'\');
          end;
          flg := FindNext(sr);
        end;
      finally
        FindClose(sr);
      end;
      for flg := 0 to fl.Count-1 do
      begin
        ALSetCdr(result, ALCons(ALNewStrAtom(fl[flg]), ALNil()));
        result := ALCdr(result);
      end;
      fl.Clear();
    end;

    if not excFl then
    begin
      flg := FindFirst(ps+ms, atr and (not faDirectory), sr);
      try
        while flg = 0 do
        begin
          if (sr.attr and faDirectory) = 0 then fl.Append(sr.name);
          flg := FindNext(sr);
        end;
      finally
        FindClose(sr);
      end;

      for flg := 0 to fl.Count-1 do
      begin
        ALSetCdr(result, ALCons(ALNewStrAtom(fl[flg]), ALNil()));
        result := ALCdr(result);
      end;
    end;

    result := ALCdr(ALPopCells(1));
  finally
    fl.Free();
  end;
end;


type
  PFilesRec = ^TFilesRec;
  TFilesRec = record
    sr: TSearchRec;
    cmode: PChar;
      // [0] is 'd', 'D' or ' '
      // [5] is always #0!
  end;


function CompFNs (fn0, fn1: string; ctype: Integer): Integer;
// ctype: 0: name & ext
// ctype: 1: name w/o ext
// ctype: 2: ext
var
  f0, f1: Integer;
begin
  if ctype = 0 then
  begin
    result := AnsiCompareText(fn0, fn1);
    exit;
  end;
  f0 := Length(fn0); while (f0 > 0) and (fn0[f0] <> '.') do Dec(f0);
  if f0 = 0 then f0 := Length(fn0)+1;
  f1 := Length(fn1); while (f1 > 0) and (fn1[f1] <> '.') do Dec(f1);
  if f1 = 0 then f1 := Length(fn1)+1;
  case ctype of
    1: result := AnsiCompareText(
         Copy(fn0, 1, f0-1),
         Copy(fn1, 1, f1-1));
    2: result := AnsiCompareText(
         Copy(fn0, f0+1, maxInt-100),
         Copy(fn1, f1+1, maxInt-100));
    else result := 0;
  end;
end;

function FilesCompare (item0, item1: Pointer): Integer;
var
  i0: PFilesRec absolute item0;
  i1: PFilesRec absolute item1;
  f: Integer;
begin
  result := 0;
  if (i0.sr.attr and faDirectory) <> (i1.sr.attr and faDirectory) then
  begin
    case i0.cmode[0] of
      'd':
        begin
          if (i0.sr.attr and faDirectory) <> 0 then result := -1 else result := 1;
          exit;
        end;
      'D':
        begin
          if (i0.sr.attr and faDirectory) <> 0 then result := 1 else result := -1;
          exit;
        end;
    end;
  end;
  f := 1;
  while true do
  begin
    case i0.cmode[f] of
      #0: break;
      'n': // name & ext
        begin
          result := CompFNs(i0.sr.name, i1.sr.name, 0);
          if result <> 0 then exit;
        end;
      'N': // name & ext
        begin
          result := -CompFNs(i0.sr.name, i1.sr.name, 0);
          if result <> 0 then exit;
        end;
      'e': // ext
        begin
          result := CompFNs(i0.sr.name, i1.sr.name, 2);
          if result <> 0 then exit;
        end;
      'E': // ext
        begin
          result := -CompFNs(i0.sr.name, i1.sr.name, 2);
          if result <> 0 then exit;
        end;
      'x': // name w/o ext
        begin
          result := CompFNs(i0.sr.name, i1.sr.name, 1);
          if result <> 0 then exit;
        end;
      'X': // name w/o ext
        begin
          result := -CompFNs(i0.sr.name, i1.sr.name, 1);
          if result <> 0 then exit;
        end;
      't': // time
        begin
          if LongWord(i0.sr.time) < LongWord(i1.sr.time) then result := -1
          else if LongWord(i0.sr.time) > LongWord(i1.sr.time) then result := 1;
          if result <> 0 then exit;
        end;
      'T': // time
        begin
          if LongWord(i0.sr.time) < LongWord(i1.sr.time) then result := 1
          else if LongWord(i0.sr.time) > LongWord(i1.sr.time) then result := -1;
          if result <> 0 then exit;
        end;
      's': // size
        begin
          if LongWord(i0.sr.size) < LongWord(i1.sr.size) then result := -1
          else if LongWord(i0.sr.size) > LongWord(i1.sr.size) then result := 1;
          if result <> 0 then exit;
        end;
      'S': // size
        begin
          if LongWord(i0.sr.size) < LongWord(i1.sr.size) then result := 1
          else if LongWord(i0.sr.size) > LongWord(i1.sr.size) then result := -1;
          if result <> 0 then exit;
        end;
    end;
    Inc(f);
  end;
end;

function ALPFiles (n: Integer): Integer;
// (files pathstr maskstr flagstr [sortstr])
// drops out "." and ".."
//  empty mask: "*"
//   builds list of files in the specified path with the specified mask.
//   "flags" flag-charz:
//     "d": include directories
//     "D": ONLY directories
//     "r": include read/only
//     "h": include hidden
//     "s": include system
//   sortstr charz:
//     "n": name, asc
//     "e": ext, asc
//     "s": size, asc
//     "t": time, asc
//     "d": dirs first
//     "N": name, desc
//     "E": ext, desc
//     "S": size, desc
//     "T": time, desc
//     "D": dirs last
//     you can say, for example, "ens" %-)
//     "name" means name & ext, if there's no "e"
// each list item is:
// (name size attr lastmoddate[unix])
// internal sort char: "x": name w/o ext
var
  ps, ms, fs, ss: string;
  sr: TSearchRec;
  fi: PFilesRec;
  fl: TList;
  cmode: array [0..5] of Char;
  cp, c: Integer;
  was, wasE: Boolean;
begin
  if not ALNextStrArg('DOS::FILES', n, ps, result) then exit;
  if not ALNextStrArg('DOS::FILES', n, ms, result) then exit;
  if not ALNextStrArg('DOS::FILES', n, fs, result) then exit;
  if not ALIsNil(n) then
  begin
    if not ALNextStrArg('DOS::FILES', n, ss, result) then exit;
  end
  else ss := '';

  // process sorting options
  FillChar(cmode[0], Length(cmode), 0);
  cmode[0] := ' '; cp := 1; wasE := false;
  for n := 1 to Length(ss) do
  begin
    if not (UpCase(ss[n]) in ['N','E','S','T','D']) then continue;
    if UpCase(ss[n]) = 'e' then wasE := true;
    was := false;
    for c := 0 to cp-1 do
      if UpCase(cmode[c]) = UpCase(ss[n]) then begin was := true; break; end;
    if was then continue;
    if UpCase(ss[n]) = 'D' then cmode[0] := ss[n]
    else begin cmode[cp] := ss[n]; Inc(cp); end;
  end;
  if wasE then
  begin
    for c := 1 to cp-1 do
    begin
      case cmode[c] of
        'N': cmode[c] := 'X';
        'n': cmode[c] := 'x';
      end;
    end;
  end;

  fl := TList.Create();
  try
    ps := StringReplace(ps, '/', '\', [rfReplaceAll]);
    ms := StringReplace(ms, '/', '\', [rfReplaceAll]);
    if ps <> '' then ps := IncludeTrailingPathDelimiter(ps);
    if (ms <> '') and (ms[1] = '\') then Delete(ms, 1, 1);
    ps := ps+ExtractFilePath(ms); ms := ExtractFileName(ms);
    if ms = '' then ms := '*';

    c := FindFirst(ps+'*', faAnyFile, sr);
    try
      while c = 0 do
      begin
        if ((sr.attr and faDirectory) <> 0) and
           ((sr.name = '.') or (sr.name = '..')) then was := false
        else
        begin
          if (sr.attr and faDirectory) <> 0 then
          begin
            was := (Pos('d', fs) <> 0);
            if not was then was := (Pos('D', fs) <> 0);
          end
          else
          begin
            was := (Pos('D', fs) = 0);
          end;
          if was then was := WildMatch(ms, sr.name, false);
          if was and ((sr.attr and faReadOnly) <> 0) then
          begin
            was := (Pos('r', fs) <> 0);
          end;
          if was and ((sr.attr and faHidden) <> 0) then
          begin
            was := (Pos('h', fs) <> 0);
          end;
          if was and ((sr.attr and faSysFile) <> 0) then
          begin
            was := (Pos('s', fs) <> 0);
          end;
        end;
        if was then
        begin
          fi := nil;
          try
            New(fi);
            fi.sr := sr; fi.cmode := @(cmode[0]);
            fl.Add(fi); fi := nil;
          finally
            if fi <> nil then Dispose(fi);
          end;
        end;
        c := FindNext(sr);
      end;
    finally
      FindClose(sr);
    end;

    fl.Sort(FilesCompare);

    result := ALCons(ALNil(), ALNil()); ALPushCell(result);
    for c := 0 to fl.Count-1 do
    begin
      n :=
        ALCons(ALNewStrAtom(PFilesRec(fl[c]).sr.name),
          ALCons(ALNewIntAtom(PFilesRec(fl[c]).sr.size),
            ALCons(ALNewIntAtom(PFilesRec(fl[c]).sr.attr),
              ALCons(ALNewIntAtom(DateTimeToUnix(FileDateToDateTime(PFilesRec(fl[c]).sr.time))),
                ALNil()
              )
            )
          )
        );
      ALSetCdr(result, ALCons(n, ALNil()));
      result := ALCdr(result);
    end;
    result := ALCdr(ALPopCells(1));
  finally
    for c := 0 to fl.Count-1 do Dispose(PFilesRec(fl[c]));
    fl.Free();
  end;
end;

function ALPCurrentDir (n: Integer): Integer;
// (current-dir)
begin
  result := ALNewStrAtom(GetCurrentDir());
end;

function ALPSetCurrentDir (n: Integer): Integer;
// (set-current-dir dir)
// returns success flag
var
  dn: string;
begin
  if not ALNextStrArg('DOS::SET-CURRENT-DIR', n, dn, result) then exit;
  if SetCurrentDir(dn) then result := ALTrue() else result := ALFalse();
end;

function ALPExpandName (n: Integer): Integer;
// (expand-name dir)
var
  fn: string;
begin
  if not ALNextStrArg('DOS::EXPAND-NAME', n, fn, result) then exit;
  fn := StringReplace(fn, '/', '\', [rfReplaceAll]);
  fn := ExpandFileName(fn);
  result := ALNewStrAtom(fn);
end;

function ALPExtractPath (n: Integer): Integer;
// (extract-path fn)
var
  fn: string;
begin
  if not ALNextStrArg('DOS::EXTRACT-PATH', n, fn, result) then exit;
  fn := StringReplace(fn, '/', '\', [rfReplaceAll]);
  fn := ExtractFilePath(fn);
  result := ALNewStrAtom(fn);
end;

function ALPExtractDir (n: Integer): Integer;
// (extract-dir fn)
var
  fn: string;
begin
  if not ALNextStrArg('DOS::EXTRACT-DIR', n, fn, result) then exit;
  fn := StringReplace(fn, '/', '\', [rfReplaceAll]);
  fn := ExtractFileDir(fn);
  result := ALNewStrAtom(fn);
end;

function ALPExtractName (n: Integer): Integer;
// (extract-name fn)
var
  fn: string;
begin
  if not ALNextStrArg('DOS::EXTRACT-NAME', n, fn, result) then exit;
  fn := StringReplace(fn, '/', '\', [rfReplaceAll]);
  fn := ExtractFileName(fn);
  result := ALNewStrAtom(fn);
end;

function ALPExtractExt (n: Integer): Integer;
// (extract-ext fn)
var
  fn: string;
begin
  if not ALNextStrArg('DOS::EXTRACT-EXT', n, fn, result) then exit;
  fn := StringReplace(fn, '/', '\', [rfReplaceAll]);
  fn := ExtractFileExt(fn);
  result := ALNewStrAtom(fn);
end;

function ALPSetExt (n: Integer): Integer;
// (set-ext fn newext [only-empty])
var
  fn, ext: string;
  onlye: Boolean;
begin
  if not ALNextStrArg('DOS::SET-EXT', n, fn, result) then exit;
  if not ALNextStrArg('DOS::SET-EXT', n, ext, result) then exit;
  n := ALCar(n);
  onlye := not ALIsNil(n) and not ALIsFalse(n);
  fn := StringReplace(fn, '/', '\', [rfReplaceAll]);
  if onlye then
  begin
    if ExtractFileExt(fn) = '' then fn := ChangeFileExt(fn, ext);
  end
  else fn := ChangeFileExt(fn, ext);
  result := ALNewStrAtom(fn);
end;

function ALPQuit (n: Integer): Integer;
// (quit [code])
var
  code: Integer;
begin
  if ALIsNil(n) then code := 0
  else if not ALNextIntArg('DOS::QUIT', n, code, result) then exit;
  //result := ALTrue();
  //ExitProcess(code);
  result := ALNewIntAtom(code);
  ALSetBreakFlag(ALBREAK_QUIT);
end;

function ALPSleep (n: Integer): Integer;
// (sleep ms)
var
  ms: Integer;
begin
  if not ALNextIntArg('DOS::SLEEP', n, ms, result) then exit;
  result := ALTrue();
  Sleep(ms);
end;

function ALPUnpackDate (n: Integer): Integer;
// (unpack-date unixtime)
// result: (y m d)
var
  ut: Integer;
  ui: Int64;
  dt: TDateTime;
  year, month, day: Word;
begin
  if not ALNextIntArg('DOS::UNPACK-DATE', n, ut, result) then exit;
  ui := LongWord(ut);
  dt := UnixToDateTime(ui);
  DecodeDate(dt, year, month, day);
  result := ALCons(ALNewIntAtom(year),
    ALCons(ALNewIntAtom(month),
      ALCons(ALNewIntAtom(day), ALNil())
    )
  );
end;

function ALPUnpackTime (n: Integer): Integer;
// (unpack-time unixtime)
// result: (h m s)
var
  ut: Integer;
  ui: Int64;
  dt: TDateTime;
  hour, minute, second, ms: Word;
begin
  if not ALNextIntArg('DOS::UNPACK-TIME', n, ut, result) then exit;
  ui := LongWord(ut);
  dt := UnixToDateTime(ui);
  DecodeTime(dt, hour, minute, second, ms);
  result := ALCons(ALNewIntAtom(hour),
    ALCons(ALNewIntAtom(minute),
      ALCons(ALNewIntAtom(second), ALNil())
    )
  );
end;

function ALPNow (n: Integer): Integer;
// (now)
// result: unix time
var
  ui: Int64;
begin
  ui := DateTimeToUnix(Now());
  result := ALNewIntAtom(ui);
end;

function ALPShortName (n: Integer): Integer;
// (short-name name)
var
  s, r: string;
  sz: Integer;
begin
  if not ALNextStrArg('DOS::SHORT-NAME', n, s, result) then exit;
  r := ' '; sz := GetShortPathName(PChar(s), @(r[1]), 0);
  SetLength(r, sz); if sz > 0 then GetShortPathName(PChar(s), @(r[1]), sz);
  result := ALNewStrAtom(r);
end;

function ALPEnvVar (n: Integer): Integer;
// (env-var name)
var
  s: string;
begin
  if not ALNextStrArg('DOS::ENV-VAR', n, s, result) then exit;
  result := ALNewStrAtom(GetEnvironmentVariable(s));
end;

function ALPSetEnvVar (n: Integer): Integer;
// (set-env-var name value)
// returns old value
var
  s, v: string;
begin
  if not ALNextStrArg('DOS::SET-ENV-VAR', n, s, result) then exit;
  if not ALNextStrArg('DOS::SET-ENV-VAR', n, v, result) then exit;
  result := ALNewStrAtom(GetEnvironmentVariable(s));
  SetEnvironmentVariable(PChar(s), PChar(v));
end;

function ALPExec (n: Integer): Integer;
// (exec execstr [opts])
// returns exitcode or *f*
// opts chars:
//   "a": "asynchronous" run (do not wait for completion)
//   "c": allocate new console for the program
//   "h": hide program window
//   "0": idle priority
//   "1": normal priority
//   "2": high priority
var
  s, so: string;
  startupInfo: TStartupInfo;
  pInfo: TProcessInformation;
  attrs: LongWord;
begin
  if not ALNextStrArg('DOS::EXEC', n, s, result) then exit;
  if not ALIsNil(n) then
  begin
    if not ALNextStrArg('DOS::EXEC', n, so, result) then exit;
  end
  else so := '';

  attrs := CREATE_DEFAULT_ERROR_MODE;
  if Pos('2', so) <> 0 then attrs := attrs or HIGH_PRIORITY_CLASS
  else if Pos('1', so) <> 0 then attrs := attrs or NORMAL_PRIORITY_CLASS
  else if Pos('0', so) <> 0 then attrs := attrs or IDLE_PRIORITY_CLASS
  else attrs := attrs or NORMAL_PRIORITY_CLASS;

  if Pos('c', so) <> 0 then attrs := attrs or CREATE_NEW_CONSOLE;

  FillChar(startupInfo, SizeOf(TStartupInfo), #0);
  startupInfo.cb := SizeOf(TStartupInfo);
  if Pos('h', so) <> 0 then
  begin
    startupInfo.dwFlags := STARTF_USESHOWWINDOW;
    startupInfo.wShowWindow := SW_HIDE;
  end;
  startupInfo.wShowWindow := startupInfo.wShowWindow or STARTF_FORCEOFFFEEDBACK;
  if not CreateProcess(nil, PChar(s), nil, nil, false,
    attrs,
    nil, // env
    nil, // dir
    startupInfo, pInfo) then begin result := ALFalse(); exit; end;
  CloseHandle(pInfo.hThread);
  if Pos('a', so) = 0 then
  begin
    WaitForSingleObject(pInfo.hProcess, INFINITE);
    if GetExitCodeProcess(pInfo.hProcess, attrs) then
      result := ALNewIntAtom(Integer(attrs))
    else result := ALNewIntAtom(-1);
  end
  else result := ALNewIntAtom(-1);
  CloseHandle(pInfo.hProcess);
end;

function ALPArgC (n: Integer): Integer;
// (argc)
begin
  result := al3dosArgV.Count-1; if result < 0 then result := 0;
  result := ALNewIntAtom(result);
end;

function ALPArgV (n: Integer): Integer;
// (argv n)
var
  c: Integer;
  s: string;
begin
  if not ALNextIntArg('DOS::ARGV', n, c, result) then exit;
  if (c < 0) or (c >= al3dosArgV.Count) then s := ''
  else s := al3dosArgV[c];
  result := ALNewStrAtom(s);
end;


procedure ALDefineDosPrimitives ();
begin
  ALAddPrimitive('dos::file-list', ALPFileList, false, 3, 4);
  ALAddPrimitive('dos::files', ALPFiles, false, 3, 4);
  ALAddPrimitive('dos::current-dir', ALPCurrentDir, false, 0, 0);
  ALAddPrimitive('dos::set-current-dir', ALPSetCurrentDir, false, 1, 1);
  ALAddPrimitive('dos::expand-name', ALPExpandName, false, 1, 1);
  ALAddPrimitive('dos::extract-path', ALPExtractPath, false, 1, 1);
  ALAddPrimitive('dos::extract-dir', ALPExtractDir, false, 1, 1);
  ALAddPrimitive('dos::extract-name', ALPExtractName, false, 1, 1);
  ALAddPrimitive('dos::extract-ext', ALPExtractExt, false, 1, 1);
  ALAddPrimitive('dos::set-ext', ALPSetExt, false, 2, 3);
  ALAddPrimitive('dos::quit', ALPQuit, false, 0, 1);
  ALAddPrimitive('dos::sleep', ALPSleep, false, 1, 1);
  ALAddPrimitive('dos::unpack-time', ALPUnpackTime, false, 1, 1);
  ALAddPrimitive('dos::unpack-date', ALPUnpackDate, false, 1, 1);
  ALAddPrimitive('dos::now', ALPNow, false, 0, 0);
  ALAddPrimitive('dos::exec', ALPExec, false, 1, 2);
  ALAddPrimitive('dos::env-var', ALPEnvVar, false, 1, 1);
  ALAddPrimitive('dos::set-env-var', ALPSetEnvVar, false, 2, 2);
  ALAddPrimitive('dos::short-name', ALPShortName, false, 1, 1);
  ALAddPrimitive('dos::argc', ALPArgC, false, 0, 0);
  ALAddPrimitive('dos::argv', ALPArgV, false, 1, 1);
end;


initialization
  al3dosArgV := TStringList.Create();
  {$IFNDEF ALISP3_MULTIENGINE}
  ALDefineDosPrimitives();
  {$ENDIF}
finalization
  FreeAndNil(al3dosArgV);
end.
