// ALISP engine v0.4.3b
// ALISP is dedicated to Enota, the girl i'll never forget
// coded by Ketmar // Vampire Avalon (psyc://ketmar.no-ip.org/~Ketmar)
// Understanding is not required. Only obedience.
//
// This program is free software. It comes without any warranty, to
// the extent permitted by applicable law. You can redistribute it
// and/or modify it under the terms of the Do What The Fuck You Want
// To Public License, Version 2, as published by Sam Hocevar. See
// http://sam.zoy.org/wtfpl/COPYING for more details.
//
{$INCLUDE alisp3def.inc}
unit alisp3WildU;

interface

uses
  SysUtils, Classes,
  wildMatchU,
 { $IFNDEF FPC}
  regExprU,
 { $ENDIF}
  alisp3U;


procedure ALDefineWildPrimitives ();


implementation


{ $IFNDEF FPC}
procedure RESplit (const aRegExpr, aInputStr: RegExprString; aPieces: TStrings);
var
  r: TRegExpr;
  prevPos: Integer;
  s: string;
begin
  aPieces.Clear();
  r := TRegExpr.Create();
  try
    r.Expression := aRegExpr;
    prevPos := 1;
    if r.Exec(aInputStr) then
    begin
      repeat
        s := Copy(aInputStr, prevPos, r.MatchPos[0]-prevPos);
        if s <> '' then aPieces.Add(s);
        s := Copy(aInputStr, r.MatchPos[0], r.MatchLen[0]);
        if s <> '' then aPieces.Add(s);
        prevPos := r.MatchPos[0]+r.MatchLen[0];
      until not r.ExecNext;
    end;
    s := Copy(aInputStr, prevPos, maxInt-10);
    if s <> '' then aPieces.Add(s);
  finally
    r.Free();
  end;
end;
{ $ENDIF}

function ALPWildMatch (n: Integer): Integer;
// (match wild str [case-sens])
var
  ws, s: string;
  cs: Boolean;
begin
  if not ALNextStrArg('WILD::MATCH', n, ws, result) then exit;
  if not ALNextStrArg('WILD::MATCH', n, s, result) then exit;
  if not ALIsNil(n) then
  begin
    n := ALCar(ALCdr(n));
    cs := not (ALIsNil(n) or ALIsFalse(n));
  end
  else cs := false;
  if WildMatch(ws, s, cs) then result := ALTrue() else result := ALFalse();
end;

{ $IFNDEF FPC}
function ALPREQuote (n: Integer): Integer;
// (re-quote str)
var
  s: string;
begin
  if not ALNextStrArg('WILD::RE-QUOTE', n, s, result) then exit;
  result := ALNewStrAtom(QuoteRegExprMetaChars(s));
end;

function ALPREMatch (n: Integer): Integer;
// (re-match re str)
// *t* or *f*
var
  re, s: string;
begin
  if not ALNextStrArg('WILD::RE-MATCH', n, re, result) then exit;
  if not ALNextStrArg('WILD::RE-MATCH', n, s, result) then exit;
  result := ALFalse();
  try
    if ExecRegExpr(re, s) then result := ALTrue();
  except
  end;
end;

function ALPRESplit (n: Integer): Integer;
// (re-split re str)
// returns list of strings
var
  re, s: string;
  sl: TStringList;
  f: Integer;
begin
  if not ALNextStrArg('WILD::RE-SPLIT', n, re, result) then exit;
  if not ALNextStrArg('WILD::RE-SPLIT', n, s, result) then exit;
  result := ALFalse(); sl := nil;
  try
    sl := TStringList.Create();
    RESplit(re, s, sl);
    f := sl.Count; result := ALNil();
    while f > 0 do
    begin
      Dec(f);
      result := ALCons(ALNewStrAtom(sl[f]), result);
    end;
  except
  end;
  sl.Free();
end;

function ALPREReplace (n: Integer): Integer;
// (re-replace re str repstr)
var
  re, s, reps: string;
begin
  if not ALNextStrArg('WILD::RE-REPLACE', n, re, result) then exit;
  if not ALNextStrArg('WILD::RE-REPLACE', n, s, result) then exit;
  if not ALNextStrArg('WILD::RE-REPLACE', n, reps, result) then exit;
  result := ALNewStrAtom('');
  try
    result := ALNewStrAtom(ReplaceRegExpr(re, s, reps, true));
  except
  end;
end;
{ $ENDIF}

procedure ALDefineWildPrimitives ();
begin
  ALAddPrimitive('wild::match', ALPWildMatch, false, 2, 3);
  { $IFNDEF FPC}
  ALAddPrimitive('wild::re-quote', ALPREQuote, false, 1, 1);
  ALAddPrimitive('wild::re-match', ALPREMatch, false, 2, 2);
  ALAddPrimitive('wild::re-split', ALPRESplit, false, 2, 2);
  ALAddPrimitive('wild::re-replace', ALPREReplace, false, 3, 3);
  { $ENDIF}
end;


begin
  {$IFNDEF ALISP3_MULTIENGINE}
  ALDefineWildPrimitives();
  {$ENDIF}
end.
