// DarkSpecter ZX Spectum Emulator
// coded by Ketmar // Vampire Avalon (psyc://ketmar.no-ip.org/~Ketmar)
// Understanding is not required. Only obedience.
//
// This program is free software. It comes without any warranty, to
// the extent permitted by applicable law. You can redistribute it
// and/or modify it under the terms of the Do What The Fuck You Want
// To Public License, Version 2, as published by Sam Hocevar. See
// http://sam.zoy.org/wtfpl/COPYING for more details.
//
//!!! COMPLETE LISP API!
// console key binding
//
unit dspEmuBindU;

interface

uses
  Windows, Messages,  SysUtils, Classes, alisp3U, lispCon3U,
  dspKeyNamesU, dspEmuVarsU;


implementation


procedure ClearBinds ();
var
  f, c: Integer;
begin
  for f := 0 to 7 do for c := 0 to 255 do conKeyBinds[f, c] := '';
  emuAltDebugKey := 0;
  emuAltDebugShift := 0;
end;

procedure ClearZXBinds ();
var
  f, c: Integer;
begin
  for f := 0 to 255 do
  begin
    pcKeyMap[false, f].ports[0] := @emuKeybMask[0]; pcKeyMap[false, f].masks[0] := $FF;
    pcKeyMap[false, f].ports[1] := @emuKeybMask[0]; pcKeyMap[false, f].masks[1] := $FF;
    pcKeyMap[true, f].ports[0] := @emuKeybMask[0]; pcKeyMap[true, f].masks[0] := $FF;
    pcKeyMap[true, f].ports[1] := @emuKeybMask[0]; pcKeyMap[true, f].masks[1] := $FF;
  end;

  for c := 0 to 1 do
  begin
    for f := 0 to 7 do
    begin
      pcJoyMap[c, f].ports[0] := @emuKeybMask[0]; pcJoyMap[c, f].masks[0] := $FF;
      pcJoyMap[c, f].ports[1] := @emuKeybMask[0]; pcJoyMap[c, f].masks[1] := $FF;
    end;
  end;
end;


procedure CmdBindList (const cmd: string; const args: array of string);
var
  f, c: Integer;
begin
  if ConIsArgHelp(args) then
  begin
    WriteLn;
    WriteLn('show all key bindings');
    WriteLn('usage: ', cmd);
    exit;
  end;

  for f := 1 to 255 do
  begin
    for c := 0 to 7 do
    begin
      if conKeyBinds[c, f] <> '' then
      begin
        WriteLn('bind ', GetPCKeyName(f, c), ' "', conKeyBinds[c, f], '"');
      end;
      if (emuAltDebugKey = f) and (emuAltDebugShift = c) then
        WriteLn('bind ', GetPCKeyName(emuAltDebugKey, emuAltDebugShift),
        ' "alt_activate_debugger"');
    end;
  end;
end;

procedure CmdQBind (const cmd: string; const args: array of string);
var
  vk: Word;
  shifts: Byte;
  s: string;
begin
  if ConIsArgHelp(args) then
  begin
    WriteLn;
    WriteLn('show key binding.');
    WriteLn('usage: ?bind keyname');
    exit;
  end;

  s := ConGetArgStr(1, args);
  if s = '' then
  begin
    WriteLn('query what?');
    exit;
  end;

  if not PCKeyNameToCode(s, vk, shifts) then
  begin
    WriteLn('unknown key name: "', s, '"');
    exit;
  end;

  if (vk = emuAltDebugKey) and (shifts = emuAltDebugShift) then
  begin
    WriteLn(s, ': "alt_activate_debugger"');
    exit;
  end;

  if conKeyBinds[shifts, vk] <> '' then
  begin
    WriteLn(s, ': "', conKeyBinds[shifts, vk], '"');
  end
  else WriteLn(s, ' is free');
end;

procedure CmdBind (const cmd: string; const args: array of string);
var
  vk: Word;
  shifts: Byte;
  s: string;
begin
  if ConIsArgHelp(args) then
  begin
    WriteLn;
    WriteLn('set key binding.');
    WriteLn('usage: bind keyname [command]');
    WriteLn(#0#1#1'if "command" arg is not given, delete binding');
    WriteLn(#0#1#1'if "key" == "clear", clear all bindings');
    exit;
  end;

  s := ConGetArgStr(1, args);
  if s = '' then
  begin
    WriteLn('bind what?');
    exit;
  end;

  if s = 'clear' then
  begin
    ClearBinds();
    WriteLn('key bindings cleared.');
    exit;
  end;

  if not PCKeyNameToCode(s, vk, shifts) then
  begin
    WriteLn('unknown key name: "', s);
    exit;
  end;

  s := ConGetArgStr(2, args);
  if s = '' then
  begin
    conKeyBinds[shifts, vk] := '';
    if (vk = emuAltDebugKey) and (shifts = emuAltDebugShift) then
    begin
      emuAltDebugKey := 0; emuAltDebugShift := 0;
    end;
  end
  else
  begin
    if s = 'alt_activate_debugger' then
    begin
      emuAltDebugKey := vk;
      emuAltDebugShift := shifts;
      conKeyBinds[shifts, vk] := '';
    end
    else conKeyBinds[shifts, vk] := s;
  end;
end;

procedure CmdZXBind (const cmd: string; const args: array of string);
var
  vk: Word;
  shifts: Byte;
  kn, zxkn: string;
begin
  if ConIsArgHelp(args) then
  begin
    WriteLn;
    WriteLn('set zx key binding.');
    WriteLn('usage: zxbind keyname zxkeys');
    WriteLn(#0#1#1'if "key" == "clear", clear all bindings');
    exit;
  end;

  kn := ConGetArgStr(1, args);
  if kn = 'clear' then
  begin
    ClearBinds();
    WriteLn('zx key bindings cleared.');
    exit;
  end;

  zxkn := ConGetArgStr(2, args);
  if (kn = '') or (zxkn = '') then
  begin
    WriteLn('bind what?');
    exit;
  end;

  if PCJoyNameValid(kn) then
  begin
    if not BindPCJoyToZX(kn, zxkn) then
    begin
      WriteLn('invalid zx bind: "', zxkn, '"');
    end;
    exit;
  end;

  if not PCKeyNameToCode(kn, vk, shifts) then
  begin
    WriteLn('unknown key name: "', kn, '"');
    exit;
  end;

  if not BindPCKeyToZX(vk, shifts, zxkn) then
  begin
    WriteLn('invalid zx bind: "', zxkn, '"');
  end;
end;


function ALQBind (n: Integer): Integer;
var
  s: string;
  vk: Word;
  shifts: Byte;
begin
  if ConALIsArgHelp(n) then
  begin
    WriteLn;
    WriteLn('get key binding');
    WriteLn('usage: (bind? "keyname")');
    WriteLn('returns bind string or nil');
    result := ALTrue();
    exit;
  end;

  if not ALNextStrArg ('BIND?', n, s, result) then exit;

  result := ALNil();
  if not PCKeyNameToCode(s, vk, shifts) then exit;

  if (vk = emuAltDebugKey) and (shifts = emuAltDebugShift) then s := 'alt_activate_debugger'
  else
  begin
    if conKeyBinds[shifts, vk] <> '' then s := conKeyBinds[shifts, vk] else exit;
  end;

  result := ALNewStrAtom(s);
end;


initialization
  ClearBinds(); ClearZXBinds();

  ALAddPrimitive('bind?', ALQBind, false, 1, 1);

  RegisterConsoleCommand('bind-list', @CmdBindList);
  RegisterConsoleCommand('?bind', @CmdQBind);
  RegisterConsoleCommand('bind', @CmdBind);
  RegisterConsoleCommand('zxbind', @CmdZXBind);
end.
