// DarkSpecter ZX Spectum Emulator
// coded by Ketmar // Vampire Avalon (psyc://ketmar.no-ip.org/~Ketmar)
// Understanding is not required. Only obedience.
//
// This program is free software. It comes without any warranty, to
// the extent permitted by applicable law. You can redistribute it
// and/or modify it under the terms of the Do What The Fuck You Want
// To Public License, Version 2, as published by Sam Hocevar. See
// http://sam.zoy.org/wtfpl/COPYING for more details.
//
// common 48k & 128k memory i/o
//
unit dsp48k128kU;

interface

uses
  SysUtils,
  z80U, dspMachConstU, dspZXDrawU, dspEmuVarsU, dspPortGeneralU, dspDbgU;


function PeekBIntComm (z80: PZ80CPU; addr: Word): Byte;
function PeekBComm (z80: PZ80CPU; addr: Word): Byte;
function ContendPeekBComm (z80: PZ80CPU; addr: Word): Byte;

function PeekBExecComm (z80: PZ80CPU): Byte;
function ContendPeekBExecComm (z80: PZ80CPU): Byte;
function DbgPeekBExecComm (z80: PZ80CPU): Byte;

procedure PokeBIntComm (z80: PZ80CPU; addr: Word; b: Byte);
procedure ContendPokeBComm (z80: PZ80CPU; addr: Word; b: Byte);
procedure PokeBComm (z80: PZ80CPU; addr: Word; b: Byte);


implementation


function PeekBIntComm (z80: PZ80CPU; addr: Word): Byte;
begin
  result := emuMemory[emuPaging[(addr and $C000) shr 14], (addr and $3FFF)];
end;

function PeekBComm (z80: PZ80CPU; addr: Word): Byte;
begin
  result := emuMemory[emuPaging[(addr and $C000) shr 14], (addr and $3FFF)];
end;

function ContendPeekBComm (z80: PZ80CPU; addr: Word): Byte;
begin
  // contention
  if (addr > 16383) and (addr <= 32767) then Inc(z80.tStates, emuXULAContention[z80.tStates]);
  result := emuMemory[emuPaging[(addr and $C000) shr 14], (addr and $3FFF)];
end;

function PeekBExecComm (z80: PZ80CPU): Byte;
var
  addr: Word;
begin
  addr := z80.regPC;
  // hook tr-dos
  if config.useTRDOS then
  begin
    if (addr >= $4000) and (emuPaging[0] = 10) then
    begin
      // tr-dos stamped in. stamp it out
      emuPaging[0] := 8+((emuLastOut7FFD and $10) shr 4);
    end
    else if (addr and $FF00) = $3D00 then
    begin
      // stamp tr-dos in?
      if (z80.prevPC >= $4000) or
         ((z80.prevPC >= $386E) and ((emuLastOut7FFD and $10) <> 0)) then emuPaging[0] := 10;
    end
    else if (emuPaging[0] = 10) and (addr = $3E01) then
    begin
      // skip delays
      z80.regA := 1; z80.regC := 1;
    end;
  end;

  result := emuMemory[emuPaging[(addr and $C000) shr 14], (addr and $3FFF)];
end;

function ContendPeekBExecComm (z80: PZ80CPU): Byte;
var
  addr: Word;
begin
  addr := z80.regPC;
  //*if (z80.tStates and $01) <> 0 then Inc(z80.tStates);
  if (addr > 16383) and (addr <= 32767) then Inc(z80.tStates, emuXULAContention[z80.tStates]);

  // hook tr-dos
  if config.useTRDOS then
  begin
    if (addr >= $4000) and (emuPaging[0] = 10) then
    begin
      // tr-dos stamped in. stamp it out
      emuPaging[0] := 8+((emuLastOut7FFD and $10) shr 4);
    end
    else if (addr and $FF00) = $3D00 then
    begin
      // stamp tr-dos in?
      if (z80.prevPC >= $4000) or
         ((z80.prevPC >= $386E) and ((emuLastOut7FFD and $10) <> 0)) then emuPaging[0] := 10;
    end
    else if (emuPaging[0] = 10) and (addr = $3E01) then
    begin
      // skip delays
      z80.regA := 1; z80.regC := 1;
    end;
  end;

  result := emuMemory[emuPaging[(addr and $C000) shr 14], (addr and $3FFF)];
end;

function DbgPeekBExecComm (z80: PZ80CPU): Byte;
var
  addr: Word;
begin
  addr := z80.regPC;
  //AlEvalStr('(poke (peek $5c78) 16384)');
  //if addr mod 4 = 0 then AlEvalStr('(lprintln (+ 1 2))');
  //if addr mod 4 = 0 then AlEvalStr('(+ 1 2)');
  (*try
    AlEvalStr('(+ 1 2)');
  except {on e: Exception do}
    begin
    end;
  end;*)

  // hook tr-dos
  if config.useTRDOS then
  begin
    if (addr >= $4000) and (emuPaging[0] = 10) then
    begin
      // tr-dos stamped in. stamp it out
      emuPaging[0] := 8+((emuLastOut7FFD and $10) shr 4);
    end
    else if (addr and $FF00) = $3D00 then
    begin
      // stamp tr-dos in?
      if (z80.prevPC >= $4000) or
         ((z80.prevPC >= $386E) and ((emuLastOut7FFD and $10) <> 0)) then emuPaging[0] := 10;
    end
    else if (emuPaging[0] = 10) and (addr = $3E01) then
    begin
      // skip delays
      z80.regA := 1; z80.regC := 1;
    end;
  end;

  DoDebugger();
  //*if config.contendedMemory and ((z80.tStates and $01)) <> 0 then Inc(z80.tStates);
  if config.contendedMemory and (addr > 16383) and (addr <= 32767) then Inc(z80.tStates, emuXULAContention[z80.tStates]);

  result := emuMemory[emuPaging[(addr and $C000) shr 14], (addr and $3FFF)];
end;

procedure PokeBIntComm (z80: PZ80CPU; addr: Word; b: Byte);
var
  p: Byte;
  te: Integer;
begin
  if addr < 16384 then exit; // ROM
  p := emuPaging[(addr and $C000) shr 14];
  addr := (addr and $3FFF);

  if (emuFramesToSkip < 0) and (p = emuScreenPage) then
  begin
    te := z80.tStates;
    if (te >= emuLineTStates[0]) and (te < emuLineTStates[240]) then
      ZXExUpdateScr(te-1, @emuMemory[emuScreenPage]);
  end;
  emuMemory[p, addr] := b;
end;

procedure PokeBComm (z80: PZ80CPU; addr: Word; b: Byte);
var
  p: Byte;
  te: Integer;
begin
  if addr < 16384 then exit; // ROM
  p := emuPaging[(addr and $C000) shr 14];
  addr := (addr and $3FFF);
  //if not emuMaxSpeed and (emuFramesToSkip < 0) and (p = emuScreenPage) and config.accurateScreenUpdate then
  //if (emuFramesToSkip < 0) and (p = emuScreenPage) and config.accurateScreenUpdate then
  if (emuFramesToSkip < 0) and (p = emuScreenPage) then
  begin
    te := z80.tStates;
    if (te >= emuLineTStates[0]) and (te < emuLineTStates[240]) then
      ZXExUpdateScr(te-1, @emuMemory[emuScreenPage]);
  end;
  emuMemory[p, addr] := b;
end;

procedure ContendPokeBComm (z80: PZ80CPU; addr: Word; b: Byte);
var
  p: Byte;
  te: Integer;
begin
  if addr < 16384 then exit; // ROM
  // contention
  //???if addr <= 32767 then Inc(z80.tStates, emuXULAContention[z80.tStates]);

  p := emuPaging[(addr and $C000) shr 14];
  addr := (addr and $3FFF);
  //if not emuMaxSpeed and (emuFramesToSkip < 0) and (p = emuScreenPage) and config.accurateScreenUpdate then
  //if (emuFramesToSkip < 0) and (p = emuScreenPage) and config.accurateScreenUpdate then
  if (emuFramesToSkip < 0) and (p = emuScreenPage) then
  begin
    te := z80.tStates;
    if (te >= emuLineTStates[0]) and (te < emuLineTStates[240]) then
      ZXExUpdateScr(te-1, @emuMemory[emuScreenPage]);
  end;
  emuMemory[p, addr] := b;
end;


end.
