// DarkSpecter ZX Spectum Emulator
// coded by Ketmar // Vampire Avalon (psyc://ketmar.no-ip.org/~Ketmar)
// Understanding is not required. Only obedience.
//
// This program is free software. It comes without any warranty, to
// the extent permitted by applicable law. You can redistribute it
// and/or modify it under the terms of the Do What The Fuck You Want
// To Public License, Version 2, as published by Sam Hocevar. See
// http://sam.zoy.org/wtfpl/COPYING for more details.
//
// Speccy screen updater
//
{$R-,S-}
unit dspZXDrawU;

interface

uses
  dspVScrU, dspMachConstU;


const
  borderH = 32;
  borderV = 24;



procedure ZXUpdateLine (y: Integer; speccyScr: Pointer); // 0..319

procedure ZXExUpdateScrInit ();
procedure ZXExUpdateScr (ticksTo: LongWord; speccyScr: Pointer);


var
  emuDrawZXLineOfs: array [0..191] of Word;
  emuDrawTickScrInfo: array [0..79999] of TVSPosInfo;
    // corresponding position information for each t-state
    // this thing is used to do accurate screen updating and
    // $FF port emulation
  emuDrawRevAttr: array [0..127] of Byte;
  emuDrawVgaHNib: array [0..15, 0..127] of LongWord;
  emuDrawPixelTable: array [0..127, 0..3] of Word;

  emuLastScrUpdate: Integer = 0;
  emuFlashState: Boolean = false;
  emuBorderColor: Byte = 0;


implementation


type
  PVideoByte = ^Byte;
  PVideoLong = ^LongWord;

  TQuad = packed record
   case Byte of
    0: (b1, b2, b3, b4: Byte);
    1: (bts: packed array [0..3] of Byte);
    2: (w1, w2: Word);
    3: (q: LongWord);
  end;


procedure ZXUpdateLine (y: Integer; speccyScr: Pointer);
var
  bitmap, attrPos: Word;
  x, color, attr, data: Byte;
  value: LongWord;
  qvalue: TQuad absolute value;
  scanline, speccyAttrs, speccyPixels: PVideoByte;
  sl: ^LongWord;
begin
  if (y < 0) or (y >= 240) then exit;
  scanline := @(dspVScreenRam[239-y]);

  // draw border
  FillChar(scanline^, 320, emuBorderColor);
  Dec(y, borderV); if (y < 0) or (y > 191) then exit;

  Inc(scanline, borderH);
  // speccy pixels
  bitmap := emuDrawZXLineOfs[y]; speccyPixels := PVideoByte(Integer(speccyScr)+bitmap);
  // speccy attrs
  attrPos := 6144 or ((y and 248) shl 2);
  speccyAttrs := PVideoByte(Integer(speccyScr)+attrPos);
  sl := Pointer(scanline);
  for x := 0 to 31 do
  begin
    color := speccyAttrs^; Inc(speccyAttrs); attr := color and 127;
    if emuFlashState and ((color and $80) <> 0) then attr := emuDrawRevAttr[attr];

    data := speccyPixels^; Inc(speccyPixels);

    sl^ := emuDrawVgaHNib[(data and $F0) shr 4, attr]; Inc(sl);
    sl^ := emuDrawVgaHNib[data and $0F, attr]; Inc(sl);
  end;
end;

procedure ZXExUpdateScrInit ();
begin
  emuLastScrUpdate := 0;
end;

procedure ZXExUpdateScr (ticksTo: LongWord; speccyScr: Pointer);
// hardcoded vistual screen dimensions were used for speed reasons!
var
  f: LongWord;
  vaddr: Integer;
  brdw: LongWord;
  b, attr: Byte;
begin
  f := emuLastScrUpdate;
  if f < LongWord(emuLineTStates[0]) then f := emuLineTStates[0];
  if ticksTo >= LongWord(emuLineTStates[240]) then ticksTo := emuLineTStates[240];

  brdw := emuDrawVgaHNib[$0F, emuBorderColor];

  f := f and (not 1);
  while f <= ticksTo do
  begin
    vaddr := emuDrawTickScrInfo[f].vscrA;
    if vaddr >= 0 then
    begin
      // draw 4 pixels
      if emuDrawTickScrInfo[f].border then
      begin
        // border
        PVideoLong(vaddr)^ := brdw;
        Inc(f, 2);
      end
      else
      begin
        repeat
          // screen
          b := Byte(PChar(speccyScr)[emuDrawTickScrInfo[f].zxattr]);
          if emuFlashState and ((b and $80) <> 0) then attr := emuDrawRevAttr[b and $7F]
          else attr := b and $7F;
          b := (Byte(PChar(speccyScr)[emuDrawTickScrInfo[f].zxscr]));
          if emuDrawTickScrInfo[f].bshift < 4 then b := b shr 4 else b := b and $0F;

          PVideoLong(vaddr)^ := emuDrawVgaHNib[b, attr];
          //PVideoWord(vaddr)^ := emuDrawPixelTable[attr, b];
          Inc(f, 2);
          vaddr := emuDrawTickScrInfo[f].vscrA;
        until emuDrawTickScrInfo[f].border or (emuDrawTickScrInfo[f].bshift = 0);
      end;
    end
    else Inc(f, 2);
  end;
  emuLastScrUpdate := f;
end;


procedure InitTables ();
var
  x, y, z, ln, t: Integer;
  foreAttr, backAttr: array [0..255] of Byte;
  qtemp: TQuad;
begin
  // speccy screen line offsets
  ln := 0;
  for t := 0 to 2 do
  begin
    for y := 0 to 7 do
    begin
      for z := 0 to 7 do
      begin
        emuDrawZXLineOfs[ln] := (t shl 11)+(z shl 8)+(y shl 5);
        Inc(ln);
      end;
    end;
  end;

  // calculate the attribute pixels
  for x := 0 to 255 do
  begin
    y := x and 7; (* fore *)
    z := (x and 56) shr 3; (* back *)
    if y = 0 then y := y or 8;
    if z = 0 then z := z or 8;

    if (x and 64) > 0 then
    begin
      y := y or 8; z := z or 8;
    end;
    foreAttr[x] := y; backAttr[x] := z;
  end;

  for x := 0 to 15 do
  begin
    for y := 0 to 127 do
    begin
      if (x and 1) <> 0 then qtemp.b4 := foreAttr[y] else qtemp.b4 := backAttr[y];
      if (x and 2) <> 0 then qtemp.b3 := foreAttr[y] else qtemp.b3 := backAttr[y];
      if (x and 4) <> 0 then qtemp.b2 := foreAttr[y] else qtemp.b2 := backAttr[y];
      if (x and 8) <> 0 then qtemp.b1 := foreAttr[y] else qtemp.b1 := backAttr[y];
      emuDrawVgaHNib[x, y] := qtemp.q;
    end;
  end;

  for y := 0 to 127 do
  begin
    for x := 0 to 3 do
    begin
      if (x and 1) <> 0 then qtemp.b2 := foreAttr[y] else qtemp.b2 := backAttr[y];
      if (x and 2) <> 0 then qtemp.b1 := foreAttr[y] else qtemp.b1 := backAttr[y];
      emuDrawPixelTable[y, x] := qtemp.w1;
    end;
  end;

  for x := 0 to 127 do
  begin
    y := (x and 7) shl 3;  //!!!K
    z := (x and 56) shr 3;
    emuDrawRevAttr[x] := (x and 64) or y or z;
  end;
end;


begin
  InitTables();
end.
