// DarkSpecter ZX Spectum Emulator
// coded by Ketmar // Vampire Avalon (psyc://ketmar.no-ip.org/~Ketmar)
// Understanding is not required. Only obedience.
//
// This program is free software. It comes without any warranty, to
// the extent permitted by applicable law. You can redistribute it
// and/or modify it under the terms of the Do What The Fuck You Want
// To Public License, Version 2, as published by Sam Hocevar. See
// http://sam.zoy.org/wtfpl/COPYING for more details.
//
// virtual screen interface
// the embedded 6x8 font has russian characters (CP866)
// character sizes & screen dimensions are hardcoded!
//
{$DEFINE USE_ASM_RESCALE}
{$R-,S-}
unit dspVScrU;

interface

uses
  Windows;


{$INCLUDE dspfont.inc}
{$INCLUDE dspicons.inc}

const
  vCharWidth = 6;
  vCharHeight = 8;

  vScrWidth = 320;
  vScrHeight = 240;

  // size of the buffer to save Clip & Origin
  vClipOriginSaveBufSize = SizeOf(TRect)+SizeOf(TPoint);

  dspIconTooSlow = 0;
  dspIconDiskIO = 1;
  dspIconContention = 2;


// update and realize palette
// data is the RGB triplets
procedure VUpdatePalette (firstColor, len: Integer; const palData);
procedure VRepaint ();
procedure VRepaint2x ();

procedure VDrawCursor (x, y, wdt, hgt: Integer; const cur; c1: Byte=0; c2: Byte=15);
procedure VDrawIcon (index, x, y: Integer);

procedure VSetOrigin (const org: TPoint); overload;
procedure VSetOrigin (x, y: Integer); overload;
procedure VSetOriginX (value: Integer);
procedure VSetOriginY (value: Integer);
procedure VResetOrigin ();
function VGetOrigin (): TPoint;
function VGetOriginX (): Integer;
function VGetOriginY (): Integer;

procedure VSetClip (x0, y0, x1, y1: Integer); overload;
procedure VSetClip (const rc: TRect); overload;
procedure VSetClipLeft (value: Integer);
procedure VSetClipTop (value: Integer);
procedure VSetClipRight (value: Integer);
procedure VSetClipBottom (value: Integer);
procedure VResetClip ();
function VGetClip (): TRect;
function VGetClipLeft (): Integer;
function VGetClipRight (): Integer;
function VGetClipTop (): Integer;
function VGetClipBottom (): Integer;

// ACHTUNG! the two following procs make NO CHECKS!
procedure VSaveClipOrigin (out buf);
procedure VRestoreClipOrigin (out buf);

// the following procedures are affected by ClipRect
// and Origin
// DSP uses this set of procs to build it's unique %-)
// window interface
// coordinate space is: left to right, top to bottom
// ClipRect is NOT shifted by Origin
// ClipRect is always inside the screen

procedure VFillRect (x0, y0, wdt, hgt: Integer; fillC: Byte; frameC: Byte = 255);
procedure VHLine (x, y, wdt: Integer; fillC: Byte);
procedure VVLine (x, y, hgt: Integer; fillC: Byte);

procedure VPutPixel (x, y: Integer; color: Byte);
procedure VDrawChar (x, y: Integer; ch: Char; fc: Byte; bc: Byte = 255);
procedure VDrawStringPart (x, y: Integer; const s: string; stPos, len: Integer;
  fc: Byte; bc: Byte = 255);
procedure VDrawString (x, y: Integer; const s: string; fc: Byte; bc: Byte = 255);


function VStrOemToAnsi (const s: string): string;
function VStrAnsiToOem (const s: string): string;

function VRect (x0, y0, x1, y1: Integer): TRect;
function VPointInRect (x, y, x0, y0, wdt, hgt: Integer): Boolean;
function VIntersectRect (out VRect: TRect; const r1, r2: TRect): Boolean;

function VCtlStrLen (const s: string): Integer;
  // length of the string w/o control char sequences (#1#n)


{$IFDEF USE_ASM_RESCALE}
procedure VScale2X (); assembler;
{$ENDIF}


type
  PVideoLineByte = ^Byte;
  TDSPRGBColor = packed record
    r, g, b: Byte;
    align: Byte; // unused
  end;
  TPalBitmapInfo = packed record
    bmiHeader: TBitmapInfoHeader;
    bmiColors: array [0..255] of Word;
  end;


var
  dspVScreenRam: packed array [0..239, 0..319] of Byte;
  vScreenRam2x: packed array [0..240*2, 0..320*2-1] of Byte;
  dspFont: packed array [Char, 0..7] of Byte;
    // high 6 bits are used
  dspWinHandle: HWND;
  vsCurrentPalette: packed array [0..255] of TDSPRGBColor;

  clipRect: TRect;
  origin: TPoint;


implementation

uses
  dspDDrawU;


{$INCLUDE dsppalette.inc}
const
  stdSpeccyPalette: array [0..15] of TDSPRGBColor = (
    (r:000; g:000; b:000; align:0),
    (r:000; g:000; b:160; align:0),
    (r:220; g:000; b:000; align:0),
    (r:228; g:000; b:180; align:0),
    (r:000; g:212; b:000; align:0),
    (r:000; g:212; b:212; align:0),
    (r:208; g:208; b:000; align:0),
    (r:200; g:200; b:200; align:0),
    (r:000; g:000; b:000; align:0),
    (r:000; g:000; b:172; align:0),
    (r:240; g:000; b:000; align:0),
    (r:252; g:000; b:220; align:0),
    (r:000; g:240; b:000; align:0),
    (r:000; g:252; b:252; align:0),
    (r:252; g:252; b:000; align:0),
    (r:252; g:252; b:252; align:0));


var
  winPalHandle: HPalette;
  bmpInfo: TPalBitmapInfo;
  bmpInfo2x: TPalBitmapInfo;


procedure VSaveClipOrigin (out buf);
begin
  Move(clipRect, buf, SizeOf(clipRect));
  Move(origin, Pointer(Integer(@buf)+SizeOf(clipRect))^, SizeOf(origin));
end;

procedure VRestoreClipOrigin (out buf);
begin
  Move(buf, clipRect, SizeOf(clipRect));
  Move(Pointer(Integer(@buf)+SizeOf(clipRect))^, origin, SizeOf(origin));
end;


function VStrAnsiToOem (const s: string): string;
begin
  result := s;
  if result <> '' then
  begin
    UniqueString(result);
    CharToOemBuff(@(result[1]), @(result[1]), Length(result));
  end;
end;

function VStrOemToAnsi (const s: string): string;
begin
  result := s;
  if result <> '' then
  begin
    UniqueString(result);
    OemToCharBuff(@(result[1]), @(result[1]), Length(result));
  end;
end;

function VRect (x0, y0, x1, y1: Integer): TRect;
begin
  result.Left := x0; result.Right := x1;
  result.Top := y0; result.Bottom := y1;
end;

function VPointInRect (x, y, x0, y0, wdt, hgt: Integer): Boolean;
begin
  result := (x >= x0) and (y >= y0) and (x < x0+wdt) and (y < y0+hgt);
end;

function VIntersectRect (out VRect: TRect; const r1, r2: TRect): Boolean;
// true: empty
begin
  VRect := r1;
  if r2.Left > r1.Left then VRect.Left := r2.Left;
  if r2.Top > r1.Top then VRect.Top := r2.Top;
  if r2.Right < r1.Right then VRect.Right := r2.Right;
  if r2.Bottom < r1.Bottom then VRect.Bottom := r2.Bottom;

  result := (VRect.Right < VRect.Left) or (VRect.Bottom < VRect.Top);
  if result then FillChar(VRect, SizeOf(VRect), 0);
end;

function VCtlStrLen (const s: string): Integer;
var
  f: Integer;
begin
  f := 1; result := Length(s);
  while f <= Length(s) do
  begin
    if s[f] = #1 then begin Dec(result, 2); Inc(f); end;
    Inc(f);
  end;
end;

{$IFDEF USE_ASM_RESCALE}
(*
procedure VScale2X (); assembler;
// vsVScreenRam --> vsVScreenRam2x
asm
  pushad
  mov   esi,offset dspVScreenRam
  mov   edi,offset vScreenRam2x
  mov   ebx,vScrWidth
  mov   edx,vScrHeight
@loopy:
  mov   ecx,ebx
@loopx:
  lodsb
  mov   ah,al
  mov   [edi+vScrWidth*2],ax
  stosw
  loop  @loopx
  add   edi,vScrWidth*2
  dec   edx
  jnz   @loopy
  popad
end;
*)
procedure VScale2X (); assembler;
// vsVScreenRam --> vsVScreenRam2x
asm
  push  ebp
  push  esi
  push  edi
  push  ebx

  mov   esi,offset dspVScreenRam
  mov   edi,offset vScreenRam2x
  mov   ebp,vScrWidth/4
  mov   edx,vScrHeight
@loopy:
  mov   ecx,ebp
@loopx:
  lodsd
  mov   bl,ah
  mov   bh,bl
  shl   ebx,16
  mov   bl,al
  mov   bh,bl
  xchg  ebx,eax
  mov   [edi+vScrWidth*2],eax
  stosd
  shr   ebx,16
  mov   al,bh
  mov   ah,al
  shl   eax,16
  mov   al,bl
  mov   ah,al
  mov   [edi+vScrWidth*2],eax
  stosd
  loop  @loopx
  add   edi,vScrWidth*2
  dec   edx
  jnz   @loopy

  pop   ebx
  pop   edi
  pop   esi
  pop   ebp
end;
{$ENDIF}

procedure VRepaint ();
var
  dc: HDC;
begin
  if dspWinHandle = 0 then exit;

  if pDDraw <> nil then
  begin
    DDrawActualizeVScreen();
    DDrawUpdate();
    exit;
  end;

  dc := GetDC(dspWinHandle);
  SelectPalette(DC, winPalHandle, false); Windows.RealizePalette(DC);
  StretchDIBits(dc,
    0, 0, 320, 240, // src
    0, 0, 320, 240, // dest
    @dspVScreenRam, PBitmapInfo(@bmpInfo)^,
    DIB_PAL_COLORS, SRCCOPY);
  ReleaseDC(dspWinHandle, dc);
end;

procedure VRepaint2x ();
var
  dc: HDC;
  {$IFNDEF USE_ASM_RESCALE}
  x, y, w2: Integer;
  src, dst: PVideoLineByte;
  {$ENDIF}
begin
  if dspWinHandle = 0 then exit;

  {$IFDEF USE_ASM_RESCALE}
  VScale2X();
  {$ELSE}
  // simple and inifficient 2x magnifier %-)
  src := @dspVScreenRam; dst := @vScreenRam2x; w2 := 320*2;
  for y := 1 to 240 do
  begin
    for x := 1 to 320 do
    begin
      dst^ := src^; Inc(dst);
      dst^ := src^; Inc(dst);
      Inc(src);
    end;
    Move(PVideoLineByte(Integer(dst)-w2)^, dst^, w2);
    Inc(dst, w2);
  end;
  {$ENDIF}

  if pDDraw <> nil then
  begin
    DDrawActualizeVScreen2x();
    DDrawUpdate();
    exit;
  end;

  dc := GetDC(dspWinHandle);
  SelectPalette(DC, winPalHandle, false); Windows.RealizePalette(DC);
  StretchDIBits(dc,
    0, 0, 640, 480, // src
    0, 0, 640, 480, // dest
    @vScreenRam2x,
    PBitmapInfo(@bmpInfo2x)^,
    DIB_PAL_COLORS, SRCCOPY);
  ReleaseDC(dspWinHandle, dc);
end;

procedure VUpdatePalette (firstColor, len: Integer; const palData);
var
  col: PLogPalette;
  f, lc: Integer;
  pc: PChar;
begin
  if firstColor < 0 then begin Inc(len, firstColor); firstColor := 0; end;
  lc := firstColor; if len > 0 then Inc(lc, len-1);
  pc := @palData;
  for f := firstColor to lc do
  begin
    Move(pc^, vsCurrentPalette[f], 4); Inc(pc, 4);
  end;

  if winPalHandle <> 0 then DeleteObject(winPalHandle); winPalHandle := 0;
  GetMem(col, SizeOf(TLogPalette)+SizeOf(TPaletteEntry)*256);
  try
    col.palVersion := $300;
    col.palNumEntries := 256;
    for f := 0 to 255 do
    begin
      col.palPalEntry[f].peRed := vsCurrentPalette[f].r;
      col.palPalEntry[f].peGreen := vsCurrentPalette[f].g;
      col.palPalEntry[f].peBlue := vsCurrentPalette[f].b;
      col.palPalEntry[f].peFlags := PC_NOCOLLAPSE;
    end;
    winPalHandle := CreatePalette(col^);
    ASSERT(winPalHandle <> 0);
  finally
    FreeMem(col);
  end;
end;

procedure VSetOrigin (const org: TPoint); overload;
begin
  origin := org;
end;

procedure VSetOrigin (x, y: Integer); overload;
begin
  origin.x := x; origin.y := y;
end;

procedure VSetOriginX (value: Integer);
begin
  origin.x := value;
end;

procedure VSetOriginY (value: Integer);
begin
  origin.x := value;
end;

procedure VResetOrigin ();
begin
  origin.x := 0; origin.y := 0;
end;

function VGetOriginX (): Integer;
begin
  result := origin.x;
end;

function VGetOriginY (): Integer;
begin
  result := origin.y;
end;

function VGetOrigin (): TPoint;
begin
  result := origin;
end;

procedure VSetClip (x0, y0, x1, y1: Integer); overload;
begin
  clipRect.Left := x0; clipRect.Right := x1;
  clipRect.Top := y0; clipRect.Bottom := y1;

  if clipRect.Left < 0 then clipRect.Left := 0
  else if clipRect.Left >= 320 then clipRect.Left := 320-1;
  if clipRect.Right < 0 then clipRect.Right := 0
  else if clipRect.Right >= 320 then clipRect.Right := 320-1;
  if clipRect.Right < clipRect.Left then clipRect.Right := clipRect.Left;

  if clipRect.Top < 0 then clipRect.Top := 0
  else if clipRect.Top >= 240 then clipRect.Top := 240-1;
  if clipRect.Bottom < 0 then clipRect.Bottom := 0
  else if clipRect.Bottom >= 240 then clipRect.Bottom := 240-1;
  if clipRect.Bottom < clipRect.Top then clipRect.Bottom := clipRect.Top;
end;

procedure VSetClip (const rc: TRect); overload;
begin
  VSetClip(rc.Left, rc.Top, rc.Right, rc.Bottom);
end;

procedure VSetClipLeft (value: Integer);
begin
  VSetClip(value, clipRect.Top, clipRect.Right, clipRect.Bottom);
end;

procedure VSetClipTop (value: Integer);
begin
  VSetClip(clipRect.Left, value, clipRect.Right, clipRect.Bottom);
end;

procedure VSetClipRight (value: Integer);
begin
  VSetClip(clipRect.Left, clipRect.Top, value, clipRect.Bottom);
end;

procedure VSetClipBottom (value: Integer);
begin
  VSetClip(clipRect.Left, clipRect.Top, clipRect.Right, value);
end;

procedure VResetClip ();
begin
  clipRect.Left := 0; clipRect.Right := 319;
  clipRect.Top := 0; clipRect.Bottom := 239;
end;

function VGetClip (): TRect;
begin
  result := clipRect;
end;

function VGetClipLeft (): Integer;
begin
  result := clipRect.Left-origin.x;
end;

function VGetClipRight (): Integer;
begin
  result := clipRect.Right-origin.x;
end;

function VGetClipTop (): Integer;
begin
  result := clipRect.Top-origin.y;
end;

function VGetClipBottom (): Integer;
begin
  result := clipRect.Bottom-origin.y;
end;

procedure VPutPixel (x, y: Integer; color: Byte);
begin
  Inc(x, origin.x); Inc(y, origin.y);
  if (color = 255) or (x < clipRect.Left) or (y < clipRect.Top) or
     (x > clipRect.Right) or (y > clipRect.Bottom) then exit;
  dspVScreenRam[239-y, x] := color;
end;

procedure VHLine (x, y, wdt: Integer; fillC: Byte);
var
  x1: Integer;
begin
  if fillC = 255 then exit;
  Inc(x, origin.x); Inc(y, origin.y);
  if (y < clipRect.Top) or (y > clipRect.Bottom) then exit;

  x1 := x+wdt-1;
  if x < clipRect.Left then x := clipRect.Left
  else if x > clipRect.Right then exit;
  if x1 < clipRect.Left then x1 := clipRect.Left
  else if x1 > clipRect.Right then x1 := clipRect.Right;
  if x1 < x then exit;

  FillChar(dspVScreenRam[239-y, x], x1-x+1, fillC);
end;

procedure VVLine (x, y, hgt: Integer; fillC: Byte);
var
  f, y1: Integer;
  wad: PVideoLineByte;
begin
  if fillC = 255 then exit;
  Inc(x, origin.x); Inc(y, origin.y);
  if (x < clipRect.Left) or (x > clipRect.Right) then exit;

  y1 := y+hgt-1;
  if y < clipRect.Top then y := clipRect.Top
  else if y > clipRect.Bottom then exit;
  if y1 < clipRect.Top then y1 := clipRect.Top
  else if y1 > clipRect.Bottom then y1 := clipRect.Bottom;
  if y1 < y then exit;

  wad := @(dspVScreenRam[239-y, x]);
  for f := y to y1 do begin wad^ := fillC; Dec(wad, 320); end;
end;

procedure VFillRect (x0, y0, wdt, hgt: Integer; fillC: Byte; frameC: Byte = 255);
var
  y, x1, y1: Integer;
  wad: PVideoLineByte;
begin
  if frameC <> 255 then
  begin
    // draw frame
    VHLine(x0, y0, wdt, frameC);
    VHLine(x0, y0+hgt-1, wdt, frameC);
    VVLine(x0, y0, hgt, frameC);
    VVLine(x0+wdt-1, y0, hgt, frameC);
    // inflate rect
    Inc(x0); Inc(y0); Dec(wdt, 2); Dec(hgt, 2);
  end;

  if fillC = 255 then exit;

  Inc(x0, origin.x); Inc(y0, origin.y);
  x1 := x0+wdt-1; y1 := y0+hgt-1;
  if x0 < clipRect.Left then x0 := clipRect.Left
  else if x0 > clipRect.Right then exit;
  if x1 < clipRect.Left then x1 := clipRect.Left
  else if x1 > clipRect.Right then x1 := clipRect.Right;
  if x1 < x0 then exit;

  if y0 < clipRect.Top then y0 := clipRect.Top
  else if y0 > clipRect.Bottom then exit;
  if y1 < clipRect.Top then y1 := clipRect.Top
  else if y1 > clipRect.Bottom then y1 := clipRect.Bottom;
  if y1 < y0 then exit;

  wdt := x1-x0+1;
  if wdt > 0 then
  begin
    // fill rect
    wad := @(dspVScreenRam[239-y0, x0]);
    for y := y0 to y1 do
    begin
      FillChar(wad^, wdt, fillC); Dec(wad, 320);
    end;
  end;
end;

procedure VDrawChar (x, y: Integer; ch: Char; fc: Byte; bc: Byte = 255);
var
  dx, dy, b, cx, w, h: Integer;
  wad: PVideoLineByte;
  c: Byte;
begin
  Inc(x, origin.x); Inc(y, origin.y);
  if (y <= clipRect.Top-8) or (y > clipRect.Bottom) or
     (x <= clipRect.Left-6) or (x > clipRect.Right) then exit; // clipped

  if y < clipRect.Top then h := 7
  else
  begin
    h := clipRect.Bottom-y; if h > 7 then h := 7;
  end;
  for dy := 0 to h do
  begin
    if y >= clipRect.Top then
    begin
      wad := @(dspVScreenRam[239-y]); Inc(wad, x);
      if x >= 0 then
      begin
        w := clipRect.Right-x; if w > 5 then w := 5;
      end
      else w := 5;
      b := dspFont[ch, dy]; cx := x;
      for dx := 0 to w do
      begin
        if cx >= clipRect.Left then
        begin
          if (b and $80) = 0 then c := bc else c := fc;
          if c <> 255 then wad^ := c;
        end;
        Inc(wad);
        b := (b and $7F) shl 1;
        Inc(cx);
      end;
    end;
    Inc(y);
  end;
end;

procedure VDrawStringPart (x, y: Integer; const s: string; stPos, len: Integer;
  fc: Byte; bc: Byte = 255);
var
  f, w: Integer;
  wad: PVideoLineByte;
  ch: Char;
  dx, dy, b, c: Integer;
  realLen: Integer;
begin
  if (fc = 255) and (bc = 255) then exit;
  if stPos < 1 then stPos := 1;
  realLen := Length(s)-stPos+1;
  if (realLen < 1) or (len < 1) then exit;
  if realLen > len then realLen := len;

  Inc(x, origin.x); Inc(y, origin.y);
  if (y <= clipRect.Top-8) or (y > clipRect.Bottom) or
     (x <= clipRect.Left-realLen*6) or (x > clipRect.Right) then exit; // clipped

  w := x-clipRect.Left;
  if w < 0 then
  begin
    w := (-w) div 6; // chars to skip
    Inc(x, w*6);
    Inc(stPos, w); Dec(realLen, w); if realLen < 1 then exit;
    if x < clipRect.Left then
    begin
      // draw partially visible char
      VDrawChar(x-origin.x, y-origin.y, s[stPos], fc, bc);
      Dec(realLen); if realLen < 1 then exit;
      Inc(x, 6); Inc(stPos);
    end;
  end;

  if x+realLen*6-1 > clipRect.Right then
  begin
    w := (clipRect.Right-x+1) div 6;
    // draw partially visible char
    VDrawChar(x-origin.x+w*6, y-origin.y, s[stPos+w], fc, bc);
    if w < 1 then exit;
    realLen := w;
  end;

  Inc(realLen, stPos-1);
  if (y < clipRect.Top) or (y > clipRect.Bottom-7) then
  begin
    // slow everything %-)
    for f := stPos to realLen do
    begin
      VDrawChar(x-origin.x, y-origin.y, s[f], fc, bc);
      Inc(x, 6);
    end;
    exit;
  end;

  for f := stPos to realLen do
  begin
    ch := s[f];
    // all partially visible chars already drawn, so
    // draw fast
    wad := @(dspVScreenRam[239-y, x]);
    for dy := 0 to 7 do
    begin
      b := dspFont[ch, dy];
      for dx := 0 to 5 do
      begin
        if (b and $80) = 0 then c := bc else c := fc;
        if c <> 255 then wad^ := c;
        b := (b and $7F) shl 1;
        Inc(wad);
      end;
      Dec(wad, 326);
    end;

    Inc(x, 6);
  end;
end;

procedure VDrawString (x, y: Integer; const s: string; fc: Byte; bc: Byte = 255);
begin
  VDrawStringPart(x, y, s, 1, Length(s), fc, bc);
end;

procedure VPrintCtlString (x, y: Integer; const s: string; const colors: array of Byte);
// color #1: default; colors counting is zero-based.
var
  attr: Word;
  ps, pe, w: Integer;
begin
  if Length(colors) < 1 then exit;
  if Length(colors) = 1 then attr := colors[0] else attr := colors[1];

  ps := 1;
  while ps <= Length(s) do
  begin
    pe := ps; while (pe <= Length(s)) and (s[pe] <> #1) do Inc(pe);
    if pe-ps > 0 then
    begin
      w := pe-ps;
      VDrawStringPart(x, y, s, ps, w, attr, 255);
      Inc(x, w*6);
    end;
    if pe >= Length(s) then break;
    ps := pe+2;
    if Byte(s[pe+1]) < Length(colors) then attr := colors[Byte(s[pe+1])]
    else attr := colors[1];
  end;
end;

procedure VDrawCursor (x, y, wdt, hgt: Integer; const cur; c1: Byte=0; c2: Byte=15);
var
  dx, dy: Integer;
  pc: PChar;
begin
  pc := @cur;
  for dy := y to y+hgt-1 do
  begin
    for dx := x to x+wdt-1 do
    begin
      if (dx >= 0) and (dy >= 0) and (dx <= 319) and (dy <= 239) then
        case pc^ of
          #1: dspVScreenRam[239-dy, dx] := c1;
          #2: dspVScreenRam[239-dy, dx] := c2;
        end;
      Inc(pc);
    end;
  end;
end;

procedure VDrawIcon (index, x, y: Integer);
var
  wdt, hgt, dx, dy: Integer;
  pc: PChar;
begin
  if (index < 0) or (index > 2) then exit;
  case index of
    0: pc := @dspIconTurtle;
    1: pc := @dspIconDisk;
    else pc := @dspIconRam;
  end;
  wdt := Byte(pc^); Inc(pc); hgt := Byte(pc^); Inc(pc);
  for dy := y to y+hgt-1 do
  begin
    for dx := x to x+wdt-1 do
    begin
      if (dx >= 0) and (dy >= 0) and (dx <= 319) and (dy <= 239) then
        dspVScreenRam[239-dy, dx] := Byte(pc^);
      Inc(pc);
    end;
  end;
end;


procedure Init ();
var
  f: Integer;
begin
  if winPalHandle <> 0 then DeleteObject(winPalHandle); winPalHandle := 0;

  VResetClip(); VResetOrigin();
  FillChar(dspVScreenRam, SizeOf(dspVScreenRam), 0);
  Move(dspStdFont, dspFont, 2048);
  FillChar(vsCurrentPalette, SizeOf(vsCurrentPalette), 0);

  bmpInfo.bmiHeader.biSize := SizeOf(bmpInfo.bmiHeader);
  bmpInfo.bmiHeader.biWidth := 320;
  bmpInfo.bmiHeader.biHeight := 240;
  bmpInfo.bmiHeader.biPlanes := 1;
  bmpInfo.bmiHeader.biBitcount := 8;
  bmpInfo.bmiHeader.biCompression := BI_RGB;
  bmpInfo.bmiHeader.biSizeimage := 0;
  bmpInfo.bmiHeader.biXpelspermeter := 0;
  bmpInfo.bmiHeader.biYpelspermeter := 0;
  bmpInfo.bmiHeader.biClrused := 255;
  bmpInfo.bmiHeader.biClrimportant := 255;

  bmpInfo2x.bmiHeader.biSize := SizeOf(bmpInfo2x.bmiHeader);
  bmpInfo2x.bmiHeader.biWidth := 640;
  bmpInfo2x.bmiHeader.biHeight := 480;
  bmpInfo2x.bmiHeader.biPlanes := 1;
  bmpInfo2x.bmiHeader.biBitcount := 8;
  bmpInfo2x.bmiHeader.biCompression := BI_RGB;
  bmpInfo2x.bmiHeader.biSizeimage := 0;
  bmpInfo2x.bmiHeader.biXpelspermeter := 0;
  bmpInfo2x.bmiHeader.biYpelspermeter := 0;
  bmpInfo2x.bmiHeader.biClrused := 255;
  bmpInfo2x.bmiHeader.biClrimportant := 255;

  // setup palette
  for f := 0 to bmpInfo.bmiHeader.biClrused-1 do
  begin
    bmpInfo.bmiColors[f] := f;
    bmpInfo2x.bmiColors[f] := f;
  end;

  VUpdatePalette(0, 16, stdSpeccyPalette);
  VUpdatePalette(16, 50, dspIconsPalette);
end;


initialization
  Init();
finalization
  if winPalHandle <> 0 then DeleteObject(winPalHandle); winPalHandle := 0;
end.
